class WrongTypeException(Exception):
    '''
    This exception should be raised if an EventIO file contains
    an unxpected Object Type at the wanted position
    '''
    def __init__(self, expected, received):
        super().__init__('Expected Type: {} but received {}'.format(
            expected, received
        ))


class WrongSizeException(Exception):
    '''
    This exception should be raised if an EventIO Object
    has the wrong size
    '''
    pass



import struct


def read_from(fmt, f):
    '''
    read the struct fmt specification from file f
    Moves the current position.
    '''
    result = struct.unpack_from(
        fmt,
        f.read(struct.calcsize(fmt))
    )
    return result


def read_ints(n, f):
    ''' read n ints from file f '''
    return read_from('{:d}i'.format(n), f)


def read_from_without_position_change(fmt, f):
    ''' Read struct format and return to old cursor position '''
    position = f.tell()
    result = read_from(fmt, f)
    f.seek(position)
    return result



import warnings
import logging
import numpy as np
from collections import namedtuple

from ..base import known_objects, EventIOFile
from ..exceptions import WrongTypeException
from .objects import (
    CorsikaRunHeader,
    CorsikaTelescopeDefinition,
    CorsikaEventHeader,
    CorsikaArrayOffsets,
    CorsikaTelescopeData,
    IACTPhotons,
    IACTLayout,
    IACTTriggerTime,
    IACTPhotoElectrons,
    CorsikaEventEndBlock,
    CorsikaRunEndBlock,
    CorsikaLongitudinal,
    CorsikaInputCard,
)


known_objects.update({
    o.eventio_type: o
    for o in [
        CorsikaRunHeader,
        CorsikaTelescopeDefinition,
        CorsikaEventHeader,
        CorsikaArrayOffsets,
        CorsikaTelescopeData,
        IACTPhotons,
        IACTLayout,
        IACTTriggerTime,
        IACTPhotoElectrons,
        CorsikaEventEndBlock,
        CorsikaRunEndBlock,
        CorsikaLongitudinal,
        CorsikaInputCard,
    ]
})

log = logging.getLogger(__name__)


class IACTFile(EventIOFile):
    '''
    An Interface to access the data of a EventIO file
    as produced by the CORSIKA IACT (a.k.a. bernlohr) extension
    more easily.

    Instead of low-level access to eventio items, it provides
    direct access to telescope events and simulation settings.

    For example, it iterates over CorsikaEvent instances and
    IACTFile[n] will return the nth event in the file.

    The structure of an IACT EventIO file is assumed to be like this:

    CorsikaRunHeader
    CorsikaInputCard
    CorsikaTelescopeDefinition

    For each Event:
      CorsikaEventHeader
      CorsikaArrayOffsets
      For each reuse:
        CorsikaTelescopeData
        For each Telescope:
          IACTPhotons
      CorsikaEventEndBlock

    CorsikaRunEndBlock
    '''

    def __init__(self, path):
        super().__init__(path)

        if not isinstance(self.objects[0], CorsikaRunHeader):
            raise WrongTypeException('Object 0 is not a CORSIKA run header')
        self.header = self.objects[0].parse_data_field()

        if not isinstance(self.objects[1], CorsikaInputCard):
            raise WrongTypeException('Object 1 is not a CORSIKA input card')
        self.input_card = self.objects[1].parse_data_field()

        if not isinstance(self.objects[2], CorsikaTelescopeDefinition):
            raise WrongTypeException('Object 2 is not a CORSIKA telescope definition')
        self.n_telescopes = self.objects[2].n_telescopes
        self.telescope_positions = self.objects[2].parse_data_field()

        self._parse_events()

        if not isinstance(self.objects[-1], CorsikaRunEndBlock):
            warnings.warn(
                'Last Object is not a CORSIKA Run End Block.'
                'The file seems to be truncated.'
            )
        else:
            self.end_block = self.objects[-1].parse_data_field()

    def __repr__(self):
        return (
            '{}(\n'
            '  path={}\n'
            '  n_telescopes={}\n'
            '  n_events={}\n'
            ')'
        ).format(
            self.__class__.__name__,
            self.path,
            self.n_telescopes,
            self.n_events,
        )

    def __len__(self):
        return self.n_events

    def __iter__(self):
        for event_num in range(self.n_events):
            yield self[event_num]

    def __getitem__(self, idx):
        if idx < 0:
            idx = self.n_events - idx
        if idx >= self.n_events:
            raise ValueError(
                'Index {} is out of range for {} with {} Events'.format(
                    idx, self.__class__.__name__, self.n_events,
                ))
        return self._build_event(idx)

    def _parse_events(self):
        ''' Sort the objects of this file according to the showers they belong to '''
        self._shower_objects = []
        reuse_values = []

        for obj in self.objects:
            if isinstance(obj, CorsikaRunEndBlock):
                break

            if isinstance(obj, CorsikaEventHeader):
                self._shower_objects.append({})
                self._shower_objects[-1]['header'] = obj

            elif isinstance(obj, CorsikaEventEndBlock):
                self._shower_objects[-1]['end_block'] = obj

            elif isinstance(obj, CorsikaArrayOffsets):
                reuse_values.append(obj.n_reuses)
                self._shower_objects[-1]['array_offsets'] = obj

            elif isinstance(obj, CorsikaTelescopeData):
                if 'telescope_data' not in self._shower_objects[-1]:
                    self._shower_objects[-1]['telescope_data'] = []
                self._shower_objects[-1]['telescope_data'].append(obj)

        self.n_showers = len(self._shower_objects)

        if reuse_values:
            assert len(reuse_values) == self.n_showers
            self.reuse = True
            self.n_events = sum(reuse_values)
            self.first_event_in_shower = np.cumsum(reuse_values) - np.array(reuse_values)
        else:
            self.reuse = False
            self.n_events = len(self._shower_objects)

    def _build_event(self, event_num):
        if self.reuse:
            shower = np.where(self.first_event_in_shower <= event_num)[0][-1]
            reuse_num = event_num - self.first_event_in_shower[shower]
        else:
            shower = event_num

        objects = self._shower_objects[shower]

        array_offset = objects['array_offsets'].parse_data_field()[reuse_num]
        time_offset = objects['array_offsets'].time_offset

        photon_bunches = {}
        n_photons = []
        n_bunches = []
        for data in objects['telescope_data'][reuse_num]:
            if isinstance(data, IACTPhotons):
                photon_bunches[data.telescope] = data.parse_data_field()
                photon_bunches[data.telescope]['x']  # -= array_offset['x']
                photon_bunches[data.telescope]['y']  # -= array_offset['y']
                photon_bunches[data.telescope]['time']  # -= time_offset
                n_photons.append(data.n_photons)
                n_bunches.append(data.n_bunches)

        event = CorsikaEvent(
            header=objects['header'].parse_data_field(),
            end_block=objects['end_block'].parse_data_field(),
            photon_bunches=photon_bunches,
            time_offset=time_offset,
            x_offset=array_offset['x'],
            y_offset=array_offset['y'],
            weight=array_offset['weight'],
            event_number=event_num,
            shower=shower,
            reuse=reuse_num + 1,
            n_photons=np.array(n_photons),
            n_bunches=np.array(n_bunches),
        )

        return event


CorsikaEventTuple = namedtuple(
    'CorsikaEventTuple',
    [
        'header', 'end_block', 'photon_bunches',
        'time_offset', 'x_offset', 'y_offset', 'weight',
        'event_number', 'shower', 'reuse',
        'n_photons', 'n_bunches',
    ]
)


class CorsikaEvent(CorsikaEventTuple):
    '''
    A single event as simulated by corsika

    Members:
      event_number

      shower:
          the id of the simulated shower

      reuse:
          reuse index for this shower

      header:
        a dictionary containing the corsika event header

      end_block:
        numpy array of floats with the event end block data

      photon_bunches:
        a dictionary mapping telescope_ids to numpy
        arrays with the photon bunch data with the following colums:
          x:         x coordinate in cm
          y:         y coordinate in cm
          cx:        cosine of incident angle in x direction
          cy:        cosine of incident angle in y direction
          time:      time since first interaction in ns
          zem:       Emission height in cm above sea level
          lambda:    wavelength in nm
          scattered: indicates if the photon was scattered in the atmosphere

      time_offset:
        time from first interaction to ground in ns

      x_offset:
        array offset in x direction in cm

      y_offset:
        array offset in y direction in cm

      weight:
        weight for this offset position.
        Only different from 1 if importance sampling was used.
    '''
    def __repr__(self):
        return '{}(event_number={}, n_telescopes={}, n_photons={})'.format(
            self.__class__.__name__,
            self.event_number,
            len(self.n_bunches),
            self.n_photons,
        )



''' Methods to read in and parse the IACT EventIO object types '''
import numpy as np
from numpy.lib.recfunctions import append_fields
import struct

from ..tools import read_ints, read_from
from ..base import EventIOObject
from ..exceptions import WrongSizeException
from .parse_corsika_data import (
    parse_corsika_event_header,
    parse_corsika_run_header,
)


__all__ = [
    'CorsikaRunHeader',
    'CorsikaTelescopeDefinition',
    'CorsikaEventHeader',
    'CorsikaArrayOffsets',
    'CorsikaTelescopeData',
    'IACTPhotons',
    'IACTLayout',
    'IACTTriggerTime',
    'IACTPhotoElectrons',
    'CorsikaEventEndBlock',
    'CorsikaRunEndBlock',
    'CorsikaLongitudinal',
    'CorsikaInputCard',
]


class CorsikaRunHeader(EventIOObject):
    '''
    This object contains the corsika run header block
    '''
    eventio_type = 1200

    def parse_data_field(self):
        '''
        Read the data in this EventIOItem

        Returns a dictionary with the items of the CORSIKA run header block
        '''
        self.seek(0)
        data = self.read()
        n, = struct.unpack('i', data[:4])
        if n != 273:
            raise WrongSizeException('Expected 273 bytes, but found {}'.format(n))

        block = np.frombuffer(
            data,
            dtype=np.float32,
            count=n,
            offset=4,
        )
        return parse_corsika_run_header(block)


class CorsikaTelescopeDefinition(EventIOObject):
    '''
    This object contains the coordinates of the telescopes of the simulated array
    '''
    eventio_type = 1201

    def __init__(self, eventio_file, header, first_byte):
        super().__init__(eventio_file, header, first_byte)
        self.n_telescopes, = read_ints(1, self)

    def __len__(self):
        return self.n_telescopes

    def parse_data_field(self):
        '''
        Read the data in this EventIOItem

        Returns a structured numpy array with columns (x, y, z, r)
        with a row for each telescope
        '''
        self.seek(4)
        data = self.read()

        number_of_following_arrays = len(data) // (self.n_telescopes * 4)
        if number_of_following_arrays != 4:
            # DN: I think this cannot happen, but who knows.
            msg = 'Number_of_following_arrays is: {}'
            raise Exception(msg.format(number_of_following_arrays))

        dtype = np.dtype('float32')
        block = np.frombuffer(
            data,
            dtype=dtype,
            count=self.n_telescopes * dtype.itemsize,
        )
        tel_pos = np.core.records.fromarrays(
            block.reshape(4, self.n_telescopes),
            names=['x', 'y', 'z', 'r'],
        )

        return tel_pos


class CorsikaEventHeader(EventIOObject):
    ''' This Object contains the CORSIKA event header block '''
    eventio_type = 1202

    def parse_data_field(self):
        '''
        Read the data in this EventIOItem

        Returns a dictionary containing the keys of the CORSIKA event header block
        '''
        self.seek(0)
        data = self.read()
        n, = struct.unpack('i', data[:4])
        if n != 273:
            raise WrongSizeException('Expected 273 bytes, but found {}'.format(n))

        block = np.frombuffer(
            data,
            dtype=np.float32,
            count=n,
            offset=4,
        )

        return parse_corsika_event_header(block)


class CorsikaArrayOffsets(EventIOObject):
    eventio_type = 1203

    def __init__(self, eventio_file, header, first_byte):
        super().__init__(eventio_file, header, first_byte)
        self.n_arrays, = read_ints(1, self)
        self.time_offset, = read_from('f', self)
        self.n_reuses = self.n_arrays

    def parse_data_field(self):
        '''
        Read the data in this EventIOItem

        Returns a structured numpy array with columns (t, x, y, weight)
        with a row for each array. This object is used to store the
        array position and contains one set of coordinates for each reuse.
        '''
        self.seek(8)
        data = self.read()

        n_columns = len(data) / (self.n_arrays * 4)
        assert n_columns.is_integer()
        n_columns = int(n_columns)
        if n_columns not in (2, 3):
            # dneise: I think this cannot happen, but who knows.
            msg = 'Number of offset columns should be in 2 or 3, found {}'
            raise Exception(msg.format(n_columns))

        positions = np.frombuffer(
            data,
            dtype=np.float32,
            count=self.n_arrays * n_columns,
        ).reshape(n_columns, self.n_arrays)

        if n_columns == 3:
            weights = positions[3, :]
        else:
            weights = np.ones(self.n_arrays, dtype=np.float32)

        return np.core.records.fromarrays(
            [positions[0, :], positions[1, :], weights],
            names=['x', 'y', 'weight'],
        )


class CorsikaTelescopeData(EventIOObject):
    '''
    A container class for the photon bunches.
    Usually contains one photon bunch object (IACTPhotons)
    per simulated telescope
    '''
    eventio_type = 1204


class IACTPhotons(EventIOObject):
    '''
    This object contains the data of the simulated cherenkov photons
    for a single telescope
    '''
    eventio_type = 1205
    columns = ('x', 'y', 'cx', 'cy', 'time', 'zem', 'photons', 'lambda')

    def __init__(self, eventio_file, header, first_byte):
        super().__init__(eventio_file, header, first_byte)
        self.compact = bool(self.header.version // 1000 == 1)

        self.array, self.telescope, self.n_photons, self.n_bunches = read_from('hhfi', self)

    def __repr__(self):
        return '{}(first={}, length={}, n_bunches={})'.format(
            self.__class__.__name__,
            self.first_byte,
            self.header.length,
            self.n_bunches,
        )

    def parse_data_field(self):
        '''
        Read the data in this EventIOObject

        Returns a numpy structured array with a record for each photon
        and the following columns:
            x:         x coordinate in cm
            y:         y coordinate in cm
            cx:        cosine of incident angle in x direction
            cy:        cosine of incident angle in y direction
            time:      time since first interaction in ns
            zem:       Emission height in cm above sea level
            lambda:    wavelength in nm
            scattered: indicates if the photon was scattered in the atmosphere
        '''

        if self.compact:
            dtype = np.dtype('int16')
        else:
            dtype = np.dtype('float32')

        if self.n_bunches == 0:
            return np.array([], dtype=[(col, dtype) for col in self.columns])

        self.seek(12)
        block = np.frombuffer(
            self.read(self.n_bunches * len(self.columns) * dtype.itemsize),
            dtype=dtype,
            count=self.n_bunches * len(self.columns)
        )
        block = block.reshape(self.n_bunches, len(self.columns))

        bunches = np.core.records.fromrecords(
            block,
            names=self.columns,
        )

        if self.compact:
            bunches = bunches.astype([(c, 'float32') for c in self.columns])
            bunches['x'] *= 0.1  # now in cm
            bunches['y'] *= 0.1  # now in cm

            # if compact, cosines are scaled by a factor of 30000
            bunches['cx'] /= 30000
            bunches['cy'] /= 30000
            #   bernloehr clips in his implementation of the reader.
            #   I am not sure I really want that.
            # bunches['cx'] = bunches['cx'].clip(a_min=-1., a_max=1.)
            # bunches['cy'] = bunches['cy'].clip(a_min=-1., a_max=1.)

            bunches['time'] *= 0.1  # in nanoseconds since first interaction.
            bunches['zem'] = np.power(10., bunches['zem']*0.001)
            bunches['photons'] *= 0.01

        bunches = append_fields(
            bunches,
            data=bunches['lambda'] < 0,
            dtypes=bool,
            names='scattered',
            usemask=False,
        )
        bunches['lambda'] = np.abs(bunches['lambda'])

        return bunches


class IACTLayout(EventIOObject):
    eventio_type = 1206


class IACTTriggerTime(EventIOObject):
    eventio_type = 1207


class IACTPhotoElectrons(EventIOObject):
    eventio_type = 1208


class CorsikaEventEndBlock(EventIOObject):
    eventio_type = 1209

    def parse_data_field(self):
        self.seek(0)
        n, = read_ints(1, self)
        if n != 273:
            raise WrongSizeException('Expected 273 bytes, but found {}'.format(n))

        dtype = np.dtype('float32')
        block = np.frombuffer(
            self.read(n * dtype.itemsize),
            dtype=dtype,
            count=n,
        )

        return block


class CorsikaRunEndBlock(EventIOObject):
    ''' This Object contains the CORSIKA run end block '''
    eventio_type = 1210

    def parse_data_field(self):
        '''
        Read the data in this EventIOObject

        Returns the CORSIKA run end block as arrays of floats.
        No parsing yet, sorry. The meaning is defined in the CORSIKA
        User Guide.
        '''

        self.seek(0)
        n, = read_ints(1, self)

        dtype = np.dtype('float32')
        block = np.frombuffer(
            self.read(n * dtype.itemsize),
            dtype=dtype,
            count=n,
        )

        return block


class CorsikaLongitudinal(EventIOObject):
    ''' This Object contains the CORSIKA longitudinal shower data block '''
    eventio_type = 1211

    def __init__(self, eventio_file, header, first_byte):
        super().__init__(eventio_file, header, first_byte)
        self.longitudinal_data = self.parse_data_field()

    def __getitem__(self, idx):
        return self.longitudinal_data[idx]

    def parse_data_field(self):
        '''
        Read the data in this EventIOObject

        Returns the CORSIKA longitudinal shower data block as arrays of floats.
        No parsing yet, sorry. The meaning is defined in the CORSIKA
        User Guide.
        '''
        self.seek(0)
        n, = read_ints(1, self)

        dtype = np.dtype('float32')
        block = np.frombuffer(
            self.read(n * dtype.itemsize),
            dtype=dtype,
            count=n,
        )

        return block


class CorsikaInputCard(EventIOObject):
    ''' This Object contains the CORSIKA steering card '''
    eventio_type = 1212

    def parse_data_field(self):
        '''
        Read the data in this EventIOObject

        Returns the CORSIKA steering card as string.
        '''
        self.seek(0)
        # read number of items in the input card
        n, = read_from('i', self)

        # corsika input card is stored as null terminated strings
        data = self.read().decode()
        strings = data.split('\0')
        inputcard = [
            remove_ascii_control_characters(string)
            for string in strings
            if string
        ]
        return inputcard


def remove_ascii_control_characters(string, mapping=dict.fromkeys(range(32))):
    ''' See http://stackoverflow.com/a/4324823/3838691 '''
    return string.translate(mapping)



import numpy as np
from functools import namedtuple
from datetime import date


def float_to_date(date_float, base_year=2000):
    '''
    Convert a float of the form YYMMDD.0 to a datetime.date
    '''
    assert float(date_float).is_integer()
    year = int(round(date_float)) // 10000 + base_year
    month = int(round(date_float % 10000)) // 100
    day = int(round(date_float % 100))

    return date(year=year, month=month, day=day)


CorsikaRunHeader = namedtuple('CorsikaRunHeader', [
    'run_id',
    'date_of_begin_run',
    'corsika_version',
    'observation_levels',
    'energy_slope',
    'energy_range',
    'EGS4_flag',
    'NKG_flag',
    'hadron_cutoff_energy',
    'muon_cutoff_energy',
    'electron_cutoff_energy',
    'photon_cutoff_energy',
    'phyiscal_constants',
    'inclinded_observation_plane_x',
    'inclinded_observation_plane_y',
    'inclinded_observation_plane_z',
    'inclinded_observation_plane_theta',
    'inclinded_observation_plane_phi',
    'NSHOW',
    'CKA',
    'CETA',
    'CSTRBA',
    'scatter_range_x',
    'scatter_range_y',
    'HLAY',
    'AATM',
    'BATM',
    'CATM',
    'NFLAIN',
    'NFLDIF',
    'NFLPI0_plus_100_times_NFLPIF',
    'NFLCHE_plus_100_times_NFRAGM',
])


def parse_corsika_run_header(run_header):
    h = run_header
    n_obs_levels = int(round(h[4]))
    if (n_obs_levels < 1) or (n_obs_levels > 10):
        raise ValueError(
            'number of observation levels n must be 0 < n < 11, but is: {}'.format(h[4])
        )

    return CorsikaRunHeader(
        run_id=h[1],
        date_of_begin_run=float_to_date(h[2]),
        corsika_version=h[3],
        observation_levels=h[5:5+n_obs_levels],
        energy_slope=h[15],
        energy_range=h[16:18],
        EGS4_flag=h[18],
        NKG_flag=h[19],
        hadron_cutoff_energy=h[20],
        muon_cutoff_energy=h[21],
        electron_cutoff_energy=h[22],
        photon_cutoff_energy=h[23],
        phyiscal_constants=h[24:74],
        inclinded_observation_plane_x=h[74],
        inclinded_observation_plane_y=h[75],
        inclinded_observation_plane_z=h[76],
        inclinded_observation_plane_theta=h[77],
        inclinded_observation_plane_phi=h[78],
        NSHOW=h[93],
        CKA=h[94:134],
        CETA=h[134:139],
        CSTRBA=h[139:150],
        scatter_range_x=h[247],
        scatter_range_y=h[248],
        HLAY=h[249:254],
        AATM=h[254:259],
        BATM=h[259:264],
        CATM=h[264:269],
        NFLAIN=h[269],
        NFLDIF=h[270],
        NFLPI0_plus_100_times_NFLPIF=h[271],
        NFLCHE_plus_100_times_NFRAGM=h[272],
    )


CorsikaEventHeader = namedtuple('CorsikaEventHeader', [
    'event_id',
    'particle_id',
    'total_energy',
    'starting_altitude',
    'number_of_first_target_if_fixed',
    'first_interaction_height',
    'momentum_x',
    'momentum_y',
    'momentum_z',
    'zenith_angle',
    'azimuth_angle',
    'seed_info',
    'run_id',
    'date_of_begin_run',
    'corsika_version',
    'observation_levels',
    'energy_slope',
    'energy_range',
    'hadron_cutoff_energy',
    'muon_cutoff_energy',
    'electron_cutoff_energy',
    'photon_cutoff_energy',
    'NFLAIN',
    'NFLDIF',
    'NFLPI0',
    'NFLPIF',
    'NFLCHE',
    'NFRAGM',
    'magnetic_field_x_component',
    'magnetic_field_z_component',
    'EGS4_flag',
    'NKG_flag',
    'low_energy_hadr_model_flag',
    'high_energy_hadr_model_flag',
    'cherenkov_bitmap',
    'neutrino_flag',
    'curved_flag',
    'unix_or_mac_flag',
    'lower_theta_limit',
    'upper_theta_limit',
    'lower_phi_limit',
    'upper_phi_limit',
    'cherenkov_bunch_size',
    'number_of_cherenkov_detectors_in_x_direction',
    'number_of_cherenkov_detectors_in_y_direction',
    'spacing_of_cherenkov_detectors_in_x_direction',
    'spacing_of_cherenkov_detectors_in_y_direction',
    'length_of_each_cherenkov_detector_in_x_direction',
    'length_of_each_cherenkov_detector_in_y_direction',
    'cherenkov_output_to_dedicated_cherenkov_output_file',
    'angle_between_x_direction_and_magnetic_north',
    'additional_muon_output_flag',
    'EGS4_multiple_scattering_step_length_factor',
    'cherenkov_bandwidth_lower_end',
    'cherenkov_bandwidth_upper_end',
    'n_reuse',
    'core_location',
    'sibyll',
    'qgsjet',
    'dpmjet',
    'venus_cross_section_flag',
    'muon_multiple_scattering_flag',
    'nkg_radial_distribution_range',
    'efrcthn',
    'weight_limit_for_thinning_hadronic',
    'weight_limit_for_thinning_em_particles',
    'max_radius_for_radial_thinning',
    'viewcone_inner_angle',
    'viewcone_outer_angle',
    'transition_energy_high_low_energy_model',
    'skimming_incidence_flag',
    'altitude_of_horizontal_shower_axis',
    'starting_height',
    'charm_generation_flag',
    'hadron_origin_of_em_subshower_flag',
    'observation_level_curvature_flag',
])


def parse_corsika_event_header(event_header):
    h = event_header

    n_random_number_sequences = int(round(h[12]))
    if (n_random_number_sequences < 1) or (n_random_number_sequences > 10):
        raise ValueError(
            'Number of random seed sequences must be 0 < n < 11, but is: {}'.format(h[12])
        )
    seed_info = h[13:13 + 3 * n_random_number_sequences].reshape(
        n_random_number_sequences, 3
    )

    n_obs_levels = int(round(h[46]))
    if (n_obs_levels < 1) or (n_obs_levels > 10):
        raise ValueError(
            'number of observation levels n must be 0 < n < 11, but is: {}'.format(h[4])
        )

    n_reuse = int(h[97])
    core_x = h[98:98+n_reuse],
    core_y = h[118:118+n_reuse],

    return CorsikaEventHeader(
        event_id=int(round(h[1])),
        particle_id=int(round(h[2])),
        total_energy=h[3],
        starting_altitude=h[4],
        number_of_first_target_if_fixed=h[5],
        first_interaction_height=h[6],
        momentum_x=h[7],
        momentum_y=h[8],
        momentum_z=h[9],
        zenith_angle=h[10],
        azimuth_angle=h[11],
        seed_info=seed_info,
        run_id=int(round(h[43])),
        date_of_begin_run=float_to_date(h[44]),
        corsika_version=h[45],
        observation_levels=h[47:47 + n_obs_levels],
        energy_slope=h[57],
        energy_range=h[58:60],
        hadron_cutoff_energy=h[60],
        muon_cutoff_energy=h[61],
        electron_cutoff_energy=h[62],
        photon_cutoff_energy=h[63],
        NFLAIN=h[64],
        NFLDIF=h[65],
        NFLPI0=h[66],
        NFLPIF=h[67],
        NFLCHE=h[68],
        NFRAGM=h[69],
        magnetic_field_x_component=h[70],
        magnetic_field_z_component=h[71],
        EGS4_flag=h[72],
        NKG_flag=h[73],
        low_energy_hadr_model_flag=h[74],
        high_energy_hadr_model_flag=h[75],
        cherenkov_bitmap=hex(int(round(h[76]))),
        neutrino_flag=h[77],
        curved_flag=h[78],
        unix_or_mac_flag=h[79],
        lower_theta_limit=h[80],
        upper_theta_limit=h[81],
        lower_phi_limit=h[82],
        upper_phi_limit=h[83],
        cherenkov_bunch_size=h[84],
        number_of_cherenkov_detectors_in_x_direction=h[85],
        number_of_cherenkov_detectors_in_y_direction=h[86],
        spacing_of_cherenkov_detectors_in_x_direction=h[87],
        spacing_of_cherenkov_detectors_in_y_direction=h[88],
        length_of_each_cherenkov_detector_in_x_direction=h[89],
        length_of_each_cherenkov_detector_in_y_direction=h[90],
        cherenkov_output_to_dedicated_cherenkov_output_file=h[91],
        angle_between_x_direction_and_magnetic_north=h[92],
        additional_muon_output_flag=h[93],
        EGS4_multiple_scattering_step_length_factor=h[94],
        cherenkov_bandwidth_lower_end=h[95],
        cherenkov_bandwidth_upper_end=h[96],
        n_reuse=n_reuse,
        core_location=np.vstack((core_x, core_y)).transpose(),
        sibyll={'interaction_flag': h[138], 'cross_section_flag': h[139]},
        qgsjet={'interaction_flag': h[140], 'cross_section_flag': h[141]},
        dpmjet={'interaction_flag': h[142], 'cross_section_flag': h[143]},
        venus_cross_section_flag=h[144],
        muon_multiple_scattering_flag=h[145],
        nkg_radial_distribution_range=h[146],
        efrcthn={
            'energy fraction of thinning level hadronic': h[147],
            'energy fraction of thinning level em-particles': h[148]
        },
        weight_limit_for_thinning_hadronic=h[149],
        weight_limit_for_thinning_em_particles=h[150],
        max_radius_for_radial_thinning=h[151],
        viewcone_inner_angle=h[152],
        viewcone_outer_angle=h[152],
        transition_energy_high_low_energy_model=h[154],
        skimming_incidence_flag=h[155],
        altitude_of_horizontal_shower_axis=h[156],
        starting_height=h[157],
        charm_generation_flag=h[158],
        hadron_origin_of_em_subshower_flag=h[159],
        observation_level_curvature_flag=h[167],
    )



import struct
from collections import namedtuple
import gzip

import logging
import warnings
import io

from .exceptions import WrongTypeException
from .tools import read_from

log = logging.getLogger(__name__)

known_objects = {}

class EventIOFile:

    def __init__(self, path):
        log.info('Opening new file {}'.format(path))
        self.path = path
        self.__file = open(path, 'rb')

        if path.endswith('.gz'):
            log.info('Found gzipped file')
            self.__compfile = gzip.GzipFile(mode='r', fileobj=self.__file)
            self.__filehandle = io.BufferedReader(self.__compfile)
        else:
            log.info('Found uncompressed file')
            self.__filehandle = self.__file

        self.objects = read_all_headers(self, toplevel=True)
        log.info('File contains {} top level objects'.format(len(self.objects)))

    def __len__(self):
        return len(self.objects)

    def seek(self, position, whence=0):
        return self.__filehandle.seek(position, whence)

    def tell(self):
        return self.__filehandle.tell()

    def read(self, size=-1):
        return self.__filehandle.read(size)

    def read_from_position(self, first_byte, size):
        pos = self.__filehandle.tell()
        self.seek(first_byte)
        data = self.read(size)
        self.seek(pos)
        return data

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__file.close()

    def __getitem__(self, idx):
        return self.objects[idx]

    def __iter__(self):
        return iter(self.objects)

    def __repr__(self):
        r = '{}(path={}, objects=[\n'.format(self.__class__.__name__, self.path)

        if len(self.objects) <= 8:
            for o in self.objects:
                r += '  {}\n'.format(o)
        else:
            for o in self.objects[:4]:
                r += '  {}\n'.format(o)
            r += '\t...\n'
            for o in self.objects[-4:]:
                r += '  {}\n'.format(o)
        r += '])'
        return r


class EventIOObject:
    eventio_type = None

    def __init__(self, eventio_file, header, first_byte):
        if header.type != self.eventio_type:
            raise WrongTypeException(self.eventio_type, header.type)

        self.eventio_file = eventio_file
        self.first_byte = first_byte
        self.header = header
        self.position = 0

        self.objects = []

        if self.header.only_sub_objects:
            self.objects = read_all_headers(self, toplevel=False)

    def __getitem__(self, idx):
        return self.objects[idx]

    def parse_data_field(self):
        ''' Read the data in this field

        should return nice python objects, e.g. structured numpy arrays
        '''
        raise NotImplementedError

    def __repr__(self):
        if len(self.objects) > 0:
            subitems = ', subitems={}'.format(len(self.objects))
        else:
            subitems = ''

        return '{}(first={}, length={}{})'.format(
            self.__class__.__name__,
            self.first_byte,
            self.header.length,
            subitems,
        )

    def read(self, size=-1):
        if size == -1 or size > self.header.length - self.position:
            size = self.header.length - self.position

        data = self.eventio_file.read_from_position(
            first_byte=self.header.data_field_first_byte + self.position, size=size,
        )

        self.position += size

        return data

    def read_from_position(self, first_byte, size):
        pos = self.tell()
        self.seek(first_byte)
        data = self.read(size)
        self.seek(pos)
        return data

    def seek(self, offset, whence=0):
        if whence == 0:
            assert offset >= 0
            self.position = offset
        elif whence == 1:
            self.position += offset
        elif whence == 2:
            self.position = self.header.length + offset
        else:
            raise ValueError('invalid whence ({}, should be 0, 1 or 2)'.format(whence))
        return self.position

    def tell(self):
        return self.position


class UnknownObject(EventIOObject):
    def __init__(self, eventio_file, header, first_byte):
        self.eventio_type = header.type
        super().__init__(eventio_file, header, first_byte)

    def __repr__(self):
        first, *last = super().__repr__().split('(first')

        return '{}[{}](first'.format(
            self.__class__.__name__, self.eventio_type
        ) + ''.join(last)

SYNC_MARKER_INT_VALUE = -736130505

def parse_sync_bytes(sync):
    ''' returns the endianness as given by the sync byte '''

    int_value, = struct.unpack('<i', sync)
    if int_value == SYNC_MARKER_INT_VALUE:
        log.debug('Found Little Endian byte order')
        return '<'

    int_value, = struct.unpack('>i', sync)
    if int_value == SYNC_MARKER_INT_VALUE:
        log.debug('Found Big Endian byte order')
        return '>'

    raise ValueError(
        'Sync must be 0xD41F8A37 or 0x378A1FD4. Got: {}'.format(sync)
    )


def read_header(f, toplevel):
    if toplevel is True:
        sync = f.read(4)
        endianness = parse_sync_bytes(sync)
        level = 0
    else:
        endianness = f.header.endianness
        level = f.header.level + 1

    if endianness == '>':
        raise NotImplementedError('Big endian byte order is not supported by this reader')

    type_version_field = read_type_field(f)
    id_field = read_from('<I', f)[0]
    only_sub_objects, length = read_length_field(f)

    if type_version_field.extended:
        length += read_extension(f)

    data_field_first_byte = f.tell()
    return_value = (
        endianness,
        type_version_field.type,
        type_version_field.version,
        type_version_field.user,
        type_version_field.extended,
        only_sub_objects,
        length,
        id_field,
        data_field_first_byte,
        level,
    )

    return return_value


HeaderBase = namedtuple(
    'HeaderBase',
    [
        'endianness', 'type', 'version', 'user', 'extended',
        'only_sub_objects',  'length', 'id', 'data_field_first_byte',
        'level'
    ]
)


class ObjectHeader(HeaderBase):
    def __new__(cls, f, parent=None):
        self = super().__new__(cls, *read_header(f, parent))
        return self


def read_all_headers(eventio_file_or_object, toplevel=True):
    eventio_file_or_object.seek(0)
    objects = []
    while True:
        position = eventio_file_or_object.tell()
        try:
            header = ObjectHeader(
                eventio_file_or_object,
                toplevel,
            )
            log.debug(
                'Found header of type {} at byte {}'.format(header.type, position)
            )
            eventio_object = known_objects.get(header.type, UnknownObject)(
                eventio_file=eventio_file_or_object,
                header=header,
                first_byte=position,
            )
            objects.append(eventio_object)
            eventio_file_or_object.seek(header.length, 1)
        except ValueError:
            warnings.warn('File seems to be truncated')
            break
        except struct.error:
            break

    return objects


# The following functions perform bit magic.
# they extract some N-bit words and 1-bit 'flags' from 32bit words
# So we need '(LEN)GTH' and '(POS)ITION' to find and extract them.
# both LEN and POS are measured in bits.
# POS starts at zero of course.

TYPE_LEN = 16
TYPE_POS = 0

USER_LEN = 1
USER_POS = 16

EXTENDED_LEN = 1
EXTENDED_POS = 17

VERSION_LEN = 12
VERSION_POS = 20

ONLYSUBOBJECTS_LEN = 1
ONLYSUBOBJECTS_POS = 30

LENGTH_LEN = 30
LENGTH_POS = 0

EXTENSION_LEN = 12
EXTENSION_POS = 0


def bool_bit_from_pos(uint32_word, pos):
    return bool(uint32_word & (1 << pos))


def len_bits_from_pos(uint32_word, len, pos):
    return (uint32_word >> pos) & ((1 << len)-1)

TypeInfo = namedtuple('TypeInfo', 'type version user extended')

def read_type_field(f):
    uint32_word = read_from('<I', f)[0]
    _type = len_bits_from_pos(uint32_word, TYPE_LEN, TYPE_POS)
    user_bit = bool_bit_from_pos(uint32_word, USER_POS)
    extended = bool_bit_from_pos(uint32_word, EXTENDED_POS)
    version = len_bits_from_pos(uint32_word, VERSION_LEN, VERSION_POS)
    return TypeInfo(_type, version, user_bit, extended)


def read_length_field(f):
    uint32_word = read_from('<I', f)[0]
    only_sub_objects = bool_bit_from_pos(uint32_word, ONLYSUBOBJECTS_POS)
    length = len_bits_from_pos(uint32_word, LENGTH_LEN, LENGTH_POS)
    return only_sub_objects, length


def read_extension(f):
    uint32_word = read_from('<I', f)[0]
    extension = len_bits_from_pos(uint32_word, EXTENSION_LEN, EXTENSION_POS)
    # we push the length-extension so many bits to the left,
    # i.e. we multiply with such a high number, that we can simply
    # use the += operator further up in `ObjectHeader_from_file` to
    # combine the normal (small) length and this extension.
    extension <<= LENGTH_LEN
    return extension



from setuptools import setup

setup(
    name='eventio',
    version='0.3.0',
    description='Python read-only implementation of the EventIO file format',
    url='https://github.com/fact-project/pyeventio',
    author='Dominik Neise, Maximilian Noethe',
    author_email='neised@phys.ethz.ch',
    license='MIT',
    packages=[
        'eventio',
        'eventio.iact',
        ],
    package_data={'eventio': ['resources/*']},
    install_requires=[
        'numpy'
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest>=3.0.0'],
)
